/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_ipt.cc,v 1.52 2002/12/13 08:05:08 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "NATCompiler_ipt.hh"

#include "fwcompiler/OSConfigurator.hh"

#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/NAT.hh"
#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/Host.hh"
#include "fwbuilder/Network.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/IPv4.hh"
#include "fwbuilder/Firewall.hh"

#include "config.h"

#include <algorithm>
#include <functional>
#include <iomanip>
#include <iostream>

#include <assert.h>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

struct subnetInfo {
    Interface *iface;
    IPv4      *ipv4;
    int        nmlength;
    subnetInfo()                           { iface=NULL; ipv4=NULL; nmlength=0; }
    subnetInfo(Interface *i,IPv4 *a,int n) { iface=i;    ipv4=a;    nmlength=n; }
};


static int chain_no=0;

string NATCompiler_ipt::myPlatformName() { return "iptables"; }

string NATCompiler_ipt::getNewTmpChainName(NATRule *rule)
{
    std::ostringstream str;

    str << "ntmp" << setw(3) << setfill('0') << chain_no;
    chain_no++;

    return str.str();
}

string NATCompiler_ipt::debugPrintRule(Rule *r)
{
    NATRule *rule=NATRule::cast(r);

    return NATCompiler::debugPrintRule(rule)+
        " "+rule->getInterfaceId()+
        " "+rule->getStr("ipt_chain")+
        " "+rule->getStr("ipt_target")+
        " (type="+rule->getRuleTypeAsString()+")";
}


int NATCompiler_ipt::prolog()
{

    int n=NATCompiler::prolog();

    if ( n>0 ) {

	bool found_ext=false;
	list<FWObject*> l2=fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    if ( iface->isExt() ) found_ext=true;
	}

	if (!found_ext)
	    throw FWException(_("At least one interface should be marked as external, can not configure NAT"));
    }


    return n;
}

bool NATCompiler_ipt::VerifyRules::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementOSrc  *osrc=rule->getOSrc();  assert(osrc);
    RuleElementODst  *odst=rule->getODst();  assert(odst);
    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);

    RuleElementTSrc  *tsrc=rule->getTSrc();  assert(tsrc);
    RuleElementTDst  *tdst=rule->getTDst();  assert(tdst);
    RuleElementTSrv  *tsrv=rule->getTSrv();  assert(tsrv);

    if (tsrc->getNeg())
	throw FWException(_("Can not use negation in translated source. Rule ")+rule->getLabel());

    if (tdst->getNeg())
	throw FWException(_("Can not use negation in translated destination. Rule ")+rule->getLabel());

    if (tsrv->getNeg())
	throw FWException(_("Can not use negation in translated service. Rule ")+rule->getLabel());

    if (tsrv->size()!=1) 
	throw FWException(_("Translated service should be 'Original' or should contain single object. Rule: ")+rule->getLabel());

    if ( Group::cast( tsrv->getFirst(true) )!=NULL)
	throw FWException(_("Can not use group in translated service. Rule ")+rule->getLabel());


    if (rule->getRuleType()==NATRule::LB)
        throw FWException(_("Load balancing rules are not supported. Rule ")+rule->getLabel());


    if (rule->getRuleType()==NATRule::DNAT) {

        if ( odst->size()!=1)
            throw FWException(_("There should be no more than one object in original destination in the rule ")+rule->getLabel());

/*
 * check removed per bug report #566172  06/07/02 --vk
 *
 *
        if ( odst->isAny() )
            throw FWException("Original destination can not be 'any' in DNAT rule. Rule "+rule->getLabel());

        Address* o2=odst->getFirst(true);
        if ( ! Host::isA(o2)  && ! Firewall::isA(o2) )
            throw FWException("Original destination must be host or firewall for this type of NAT rule. Rule "+rule->getLabel());
*/

/*
 *  check if we have specific IP address to use in odst 
 *
 *  disabled to fix a bug #562173 (bug created per user's request on
 *  Open Forum) 05/29/02 --vk
 *
        bool find_ext_static_address=false;
        FWObjectTypedChildIterator j=o2->findByType(Interface::TYPENAME);
        for ( ; j!=j.end(); ++j ) {
            Interface *iface=Interface::cast(*j);
            if ( ! iface->isLoopback() && iface->isExt() && ! iface->isDyn() ) 
                find_ext_static_address=true;
        }
        if ( ! find_ext_static_address && Host::isA(o2) ) 
            find_ext_static_address=true;

        if ( ! find_ext_static_address ) {
            if ( Firewall::isA(o2) )
                throw FWException("Could not find appropriate address for original destination: external interface of the firewall has dynamic address. Rule "+rule->getLabel());
            else
                throw FWException("Could not find appropriate address for original destination. Rule "+rule->getLabel());
        }
*/
    }

    if (rule->getRuleType()==NATRule::SNAT ) {
        if ( tsrc->size()!=1)
            throw FWException(_("There should be no more than one object in translated source in the rule ")+rule->getLabel());

        Address* o1=tsrc->getFirst(true);
        if ( ! tsrc->isAny() && Network::cast(o1)!=NULL)
            throw FWException(_("Can not use network object in translated source. Rule ")+rule->getLabel());
    }


    if (rule->getRuleType()==NATRule::SNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(osrc->getFirst(true));
        Network *a2=Network::cast(tsrc->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException(_("Original and translated source should both be networks of the same size . Rule ")+rule->getLabel());
    }

    if (rule->getRuleType()==NATRule::DNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(odst->getFirst(true));
        Network *a2=Network::cast(tdst->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException(_("Original and translated destination should both be networks of the same size . Rule ")+rule->getLabel());
    }

    return true;
}

/*
 * this should be called only after splitServices, so that we have
 * objects of the same type in OSrv and either "any" or a single
 * object in TSrv
 */
bool NATCompiler_ipt::VerifyRules2::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    if (rule->getRuleType()!= NATRule::Return)
    {
        RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);
        RuleElementTSrv  *tsrv=rule->getTSrv();  assert(tsrv);

        Service *s1=osrv->getFirst(true);
        Service *s2=tsrv->getFirst(true);

        if (osrv->isAny() && ! tsrv->isAny())
            throw FWException(_("Can not use service object in Translated Service if Original Service is 'Any'. Rule ")+rule->getLabel());
        
        if (!tsrv->isAny() && s1->getProtocolNumber()!=s2->getProtocolNumber())
            throw FWException(_("Translated Service should be either 'Original' or should contain object of the same type as Original Service. Rule ")+rule->getLabel());
    }
    return true;
}



bool NATCompiler_ipt::portTranslationRules::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    Address  *odst=rule->getODst()->getFirst(true);
    Service  *osrv=rule->getOSrv()->getFirst(true);

    Address  *tsrc=rule->getTSrc()->getFirst(true);
    Address  *tdst=rule->getTDst()->getFirst(true);
    Service  *tsrv=rule->getTSrv()->getFirst(true);

    if (rule->getRuleType() == NATRule::DNAT &&
        tsrc->isAny() &&   tdst->isAny() && ! tsrv->isAny() )
    {
        rule->getTDst()->addRef( odst );
    }

    tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::specialCaseWithRedirect::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    Address  *tdst=rule->getTDst()->getFirst(true);

/* we consider rule redirect only if TDst is a firewall object */
    if (rule->getRuleType() == NATRule::DNAT && tdst->getId()==compiler->fw->getId()) 
        rule->setRuleType(NATRule::Redirect);

    tmp_queue.push_back(rule);

    return true;
}


bool NATCompiler_ipt::splitOnOSrv::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);
    if (osrv->size()!=1) {

	for(list<FWObject*>::iterator i=osrv->begin(); i!=osrv->end(); ++i) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) 
		o=FWReference::cast(o)->getPointer();
	    Service *s=Service::cast( o );
	    assert(s);

	    NATRule *r= NATRule::cast(
				      compiler->dbcopy->create(NATRule::TYPENAME,true) );
	    compiler->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementOSrv *nosrv=r->getOSrv();
	    nosrv->clearChildren();

	    nosrv->addRef( s );

	    tmp_queue.push_back( r );
	}
    } else
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::fillTranslatedSrv::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    Service  *osrv_o=rule->getOSrv()->getFirst(true);
    Service  *tsrv_o=rule->getTSrv()->getFirst(true);

    if ( ! osrv_o->isAny()  && tsrv_o->isAny() ) {
	RuleElementTSrv  *tsrv=rule->getTSrv();
	tsrv->addRef(osrv_o);
    }
    return true;
}

bool NATCompiler_ipt::addVirtualAddress::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    Address *a=NULL;

    if (rule->getRuleType()==NATRule::SNAT || rule->getRuleType()==NATRule::DNAT) 
    {
        if (rule->getRuleType()==NATRule::SNAT)   
            a=rule->getTSrc()->getFirst(true);
        else
            a=rule->getODst()->getFirst(true);

        if ( ! a->isAny() && ! compiler->complexMatch(a,compiler->fw) )
            compiler->osconfigurator->addVirtualAddressForNAT( a->getAddress() );

        return true;
    }


    if (rule->getRuleType()==NATRule::SNetnat || rule->getRuleType()==NATRule::DNetnat)
    {
        if (rule->getRuleType()==NATRule::SNetnat)   
            a=rule->getTSrc()->getFirst(true);
        else
            a=rule->getODst()->getFirst(true);

        if ( ! a->isAny() && Network::cast(a) )
            compiler->osconfigurator->addVirtualAddressForNAT( *(Network::cast(a)) );

        return true;
    }

    return true;
}

bool NATCompiler_ipt::splitRuleIfRuleElementIsDynamicInterface::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElement    *re=RuleElement::cast(rule->getFirstByType(re_type));
    int nre=re->size();

    vector<FWObject*> cl;

    for(list<FWObject*>::iterator i=re->begin(); nre>1 && i!=re->end(); ++i) {
        FWObject *o= *i;
	FWObject *obj = NULL;
        if (FWReference::cast(o)!=NULL) obj=compiler->getCachedObj(o->getStr("ref"));
        Interface *iface=Interface::cast(obj);
        if (iface!=NULL && iface->isDyn()) {

	    cl.push_back(o);   // can not remove right now because remove invalidates iterator
            nre--;

	    NATRule  *new_rule= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	    compiler->temp_ruleset->add(new_rule);
	    new_rule->duplicate(rule);
            RuleElement *new_re=RuleElement::cast(new_rule->getFirstByType(re_type));
	    new_re->clearChildren();
	    new_re->setAnyElement();
	    new_re->addRef( iface );
	    tmp_queue.push_back(new_rule);
        }
    }
    if (!cl.empty()) {
        for (vector<FWObject*>::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  
            re->remove( (*i1) );
    }

    tmp_queue.push_back(rule);

    return true;
}



void NATCompiler_ipt::specialCaseWithDynInterface::dropDynamicInterface(RuleElement *re)
{
    list<FWObject*> cl;
    for (list<FWObject*>::iterator i1=re->begin(); i1!=re->end(); ++i1) 
    {
        FWObject *o   = *i1;
        FWObject *obj = o;
        if (FWReference::cast(o)!=NULL) obj=compiler->getCachedObj(o->getStr("ref"));
        Interface  *ifs   =Interface::cast( obj );

        if (ifs!=NULL && ifs->isDyn()) continue;
        cl.push_back(obj);
    }
    if (!cl.empty()) 
    {
        re->clearChildren();
        for (list<FWObject*>::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  
            re->addRef( (*i1) );
    }
}

bool NATCompiler_ipt::specialCaseWithDynInterface::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    switch (rule->getRuleType()) {
    case NATRule::Masq:
    case NATRule::SNAT:
        dropDynamicInterface( rule->getOSrc() );
        break;
    case NATRule::DNAT:
        dropDynamicInterface( rule->getODst() );
        break;
    }

    tmp_queue.push_back(rule);
    return true;
}


/*
 *  I assume that there is always only one object in ODst, TSrc and TDst
 *  rule elements. This should have been assured by inspector VerifyRules
 */
bool NATCompiler_ipt::ReplaceFirewallObjectsODst::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    list<FWObject*> cl;
    RuleElementODst *rel;
    Address     *obj=NULL;

    switch (rule->getRuleType()) {

    case NATRule::Masq:
//    case NATRule::Redirect: 
	return true;
    default:
	rel=rule->getODst();      assert(rel);
	obj=rel->getFirst(true);  assert(obj!=NULL);

	if (obj->getId()==compiler->getFwId() ) 
	{

	    list<FWObject*> l2=compiler->fw->getByType(Interface::TYPENAME);
	    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
		Interface *interface_=Interface::cast(*i);
/*
 * Right now I build DNAT and Redirect rules only for external interface.
 * This actually implies some sort of assymmetric firewall; it does
 * not have to be like this.
 * 
 * TODO: add platform-specific fw option "Generate NAT rules for all
 * interfaces" and use here all interfaces , not only external ones.
 * Do this for v1.0.4.  Modify verification code in verifyRules
 * accordingly. The same applies to ReplaceFirewallObjectsTSrc
 *
 * Note 05/29/02: I do not check if external interface is dynamic anymore
 *                see bug #562173    --vk
 *
 *
 *		if (! interface_->isLoopback() && 
 *                    interface_->isExt() &&
 *                  ! interface_->isDyn() ) cl.push_back(interface_);
 */
		if (! interface_->isLoopback() && 
                      interface_->isExt() ) cl.push_back(interface_);
	    }
	    if ( ! cl.empty() ) {
		while (rel->size()) 
		    rel->remove( rel->front()  );
		
		for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
		{
		    rel->addRef( *i1 );
		}
	    }
	}
    }
    return true;
}

bool NATCompiler_ipt::ReplaceFirewallObjectsTSrc::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    list<FWObject*> cl;
    RuleElementTSrc *rel;
    Address     *obj=NULL;

    switch (rule->getRuleType()) {

    case NATRule::Masq:
    case NATRule::Redirect: return true;

    default:
	rel=rule->getTSrc();      assert(rel);
	obj=rel->getFirst(true);  assert(obj!=NULL);

	if (obj->getId()==compiler->getFwId() ) 
	{

	    list<FWObject*> l2=compiler->fw->getByType(Interface::TYPENAME);
	    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
		Interface *interface_=Interface::cast(*i);
/*
 *  see comment above in ReplaceFirewallObjectsODst
 *
 * Note 06/04/02: I do not check if external interface is dynamic anymore
 *                see bug #562173    --vk
 *
 *	if (! interface_->isLoopback() &&
 *                    interface_->isExt() && 
 *                  ! interface_->isDyn() ) cl.push_back(interface_);
 */
		if (! interface_->isLoopback() &&
                      interface_->isExt() )  cl.push_back(interface_);
	    }
	    if ( ! cl.empty() ) {
		while (rel->size()) 
		    rel->remove( rel->front()  );
		
		for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
		{
		    rel->addRef( *i1 );
		}
	    }
	}
    }
    return true;
}

bool NATCompiler_ipt::dynamicInterfaceInODst::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementODst  *odstrel=rule->getODst();  assert(odstrel);
    Address          *odst   =odstrel->getFirst(true);
    if ( ! odstrel->isAny() && 
	 Interface::cast(odst)!=NULL &&
	 Interface::cast(odst)->isDyn() ) {

	odstrel->clearChildren();
	odstrel->setAnyElement();
    }
    return true;
}

bool NATCompiler_ipt::dynamicInterfaceInTSrc::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    Address  *tsrc=rule->getTSrc()->getFirst(true);

    if (rule->getRuleType()==NATRule::SNAT && 
        Interface::cast(tsrc)!=NULL &&
	Interface::cast(tsrc)->isDyn()) {

            rule->setRuleType(NATRule::Masq);

            if ( rule->getStr("ipt_target")=="" || rule->getStr("ipt_target")=="SNAT" )
                rule->setStr("ipt_target","MASQUERADE");
    }
    return true;
}

/**
 * unlike standard inspector addressRanges in the base class NATCompiler,
 * this one does not expand address ranges in TSrc and TDst because
 * iptables supports ranges in those rule elements
 */
bool NATCompiler_ipt::addressRanges::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElement *rel;

    rel=rule->getOSrc();    assert(rel);
    compiler->_expandAddressRanges(rule,rel);
    rel=rule->getODst();    assert(rel);
    compiler->_expandAddressRanges(rule,rel);
    return true;
}



bool NATCompiler_ipt::splitServices::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *srv=rule->getOSrv();

    if (srv->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    map<int, list<Service*> > services;

 loop:
    for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) {
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));

	Service *s=Service::cast( o );
	assert(s);

	int proto=s->getProtocolNumber();
	services[proto].push_back(s);
    }

    for (map<int, list<Service*> >::iterator i=services.begin(); i!=services.end(); i++) {
	list<Service*> &sl=(*i).second;

	NATRule *r= NATRule::cast(
	    compiler->dbcopy->create(NATRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	RuleElementOSrv *nsrv=r->getOSrv();
	nsrv->clearChildren();

	for (list<Service*>::iterator j=sl.begin(); j!=sl.end(); j++) {
	    nsrv->addRef( (*j) );
	}

	tmp_queue.push_back(r);

    }
    return true;
}

bool NATCompiler_ipt::separatePortRanges::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *rel= rule->getOSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( TCPService::isA(s) || UDPService::isA(s) ) {
            int srs=s->getInt("src_range_start");
            int sre=s->getInt("src_range_end");
            int drs=s->getInt("dst_range_start");
            int dre=s->getInt("dst_range_end");

            compiler->normalizePortRange(srs,sre);
            compiler->normalizePortRange(drs,dre);

            if (srs!=sre || drs!=dre) {
                NATRule *r= NATRule::cast(
                    compiler->dbcopy->create(NATRule::TYPENAME,false) );
                compiler->temp_ruleset->add(r);
                r->duplicate(rule);
                RuleElementOSrv *nsrv=r->getOSrv();
                nsrv->clearChildren();
                nsrv->addRef( s );
                tmp_queue.push_back(r);
                services.push_back(s);
            } 
        }
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::separateSourcePorts::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *rel= rule->getOSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    NATRule         *rule_4_src_ports=NULL;
    RuleElementOSrv *nsrv;

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) 
    {
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( TCPService::isA(s) || UDPService::isA(s) ) {
            int srs=s->getInt("src_range_start");
            int sre=s->getInt("src_range_end");

            compiler->normalizePortRange(srs,sre);

            if (srs!=0 || sre!=0)
            {
                if (rule_4_src_ports==NULL)
                {
                    rule_4_src_ports= NATRule::cast(
                        compiler->dbcopy->create(NATRule::TYPENAME,false) );
                    compiler->temp_ruleset->add(rule_4_src_ports);
                    rule_4_src_ports->duplicate(rule);
                    nsrv=rule_4_src_ports->getOSrv();
                    nsrv->clearChildren();
                    tmp_queue.push_back(rule_4_src_ports);
                }
                nsrv->addRef( s );
                services.push_back(s);
            }
        }
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::separateSourceAndDestinationPorts::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *rel= rule->getOSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    NATRule         *nrule=NULL;
    RuleElementOSrv *nsrv;

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) 
    {
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( TCPService::isA(s) || UDPService::isA(s) ) {
            int srs=s->getInt("src_range_start");
            int sre=s->getInt("src_range_end");
            int drs=s->getInt("dst_range_start");
            int dre=s->getInt("dst_range_end");

            compiler->normalizePortRange(srs,sre);
            compiler->normalizePortRange(drs,dre);

            if ( (srs!=0 || sre!=0) && (drs!=0 || dre!=0) )
            {
                if (nrule==NULL)
                {
                    nrule= NATRule::cast(
                        compiler->dbcopy->create(NATRule::TYPENAME,false) );
                    compiler->temp_ruleset->add(nrule);
                    nrule->duplicate(rule);
                    RuleElementOSrv *nsrv=nrule->getOSrv();
                    nsrv->clearChildren();
                    tmp_queue.push_back(nrule);
                }
                nsrv->addRef( s );
                services.push_back(s);
            } 
        }
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}


bool NATCompiler_ipt::prepareForMultiport::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *rel= rule->getOSrv();
    Service         *srv= rel->getFirst(true);

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

/*
 * processor splitServices should have been called eariler, so now all
 * services in Srv are of the same type
 */
    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	rule->setBool("ipt_multiport",true);

/* TODO: make sure we have no more than 15 ports */

	tmp_queue.push_back(rule);
    }
    return true;
}

bool NATCompiler_ipt::doOSrcNegation::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrc *osrcrel=rule->getOSrc();

/*   ! A  B  C    */

    if (osrcrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=NATCompiler_ipt::getNewTmpChainName(rule);
	osrcrel->setNeg(false);
/*
 *   negation in OSrc :
 *
 *  CHAIN     !A    B    C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----      any   B    C    SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN   A   any  any   RETURN     RETURN
 * TMP_CHAIN  any  any   C   SNAT/DNAT  ---------
 */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	tmp_queue.push_back(r);

/* TMP_CHAIN   A  any  any  RETURN  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any any C  ACTION  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
//	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::doODstNegation::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementODst *odstrel=rule->getODst();

/*   ! A  B  C    */

    if (odstrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=NATCompiler_ipt::getNewTmpChainName(rule);
	odstrel->setNeg(false);
/*
 *   negation in Odst :
 *
 *  CHAIN      A    !B   C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----       A   any   C    SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN  any   B   any   RETURN     RETURN
 * TMP_CHAIN  any  any   C    SNAT/DNAT  ---------
 */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getODst();   ndst->clearChildren();   ndst->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  B  any  RETURN  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  any  C  ACTION  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
//	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::doOSrvNegation::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementOSrv *osrvrel=rule->getOSrv();

/*   A  B ! C    */

    if (osrvrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=NATCompiler_ipt::getNewTmpChainName(rule);
	osrvrel->setNeg(false);
/*
 *   negation in OSrv :
 *
 *  CHAIN      A    B   !C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----       A    B   any   SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN  any  any   C    RETURN     RETURN
 * TMP_CHAIN  any  any  any   SNAT/DNAT  ---------
 */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	tmp_queue.push_back(r);

/* TMP_CHAIN   A  B  any  RETURN  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any B  any  ACTION  */
	r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}

bool NATCompiler_ipt::splitNONATRule::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    if ( rule->getStr("ipt_chain").empty() && rule->getRuleType()==NATRule::NONAT) {

        Address *osrc=rule->getOSrc()->getFirst(true);
        if ( compiler->complexMatch(osrc,compiler->fw) ) {
            rule->setStr("ipt_chain","OUTPUT");  
            RuleElement *rel=rule->getOSrc();
	    rel->clearChildren();
	    rel->setAnyElement();
            tmp_queue.push_back(rule);
        } else {
            rule->setStr("ipt_chain","PREROUTING");  
            tmp_queue.push_back(rule);

            NATRule *r= NATRule::cast(compiler->dbcopy->create(NATRule::TYPENAME,true) );
            compiler->temp_ruleset->add(r);
            r->duplicate(rule);
            r->setStr("ipt_chain","POSTROUTING");  
            tmp_queue.push_back(r);
        }
    } else
        tmp_queue.push_back(rule);
    return true;
}


bool NATCompiler_ipt::decideOnChain::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    if ( ! rule->getStr("ipt_chain").empty() ) return true; // already defined

    switch (rule->getRuleType()) {
    case NATRule::SNAT:     rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::SNetnat:  rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::Masq:     rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::DNAT:     rule->setStr("ipt_chain","PREROUTING");  break;
    case NATRule::DNetnat:  rule->setStr("ipt_chain","PREROUTING");  break;
    case NATRule::Redirect: rule->setStr("ipt_chain","PREROUTING");  break;
    case NATRule::NONAT:    
// processor splitNONATRule took care of NONAT rule
        break;
    }
    return true;
}


bool NATCompiler_ipt::decideOnTarget::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    if ( ! rule->getStr("ipt_target").empty() ) return true; // already defined

    switch (rule->getRuleType()) {
    case NATRule::NONAT:    rule->setStr("ipt_target","ACCEPT");   break;
    case NATRule::SNAT:     rule->setStr("ipt_target","SNAT");     break;
    case NATRule::SNetnat:  rule->setStr("ipt_target","NETMAP");   break;
    case NATRule::DNAT:     rule->setStr("ipt_target","DNAT");     break;
    case NATRule::DNetnat:  rule->setStr("ipt_target","NETMAP");   break;
    case NATRule::Masq:     rule->setStr("ipt_target","MASQUERADE"); break;
    case NATRule::Redirect: rule->setStr("ipt_target","REDIRECT"); break;
    case NATRule::Return:   rule->setStr("ipt_target","RETURN");   break;
    }
    return true;
}

bool NATCompiler_ipt::AssignInterface::processNext()
{
    NATRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    Address  *a=NULL;
    FWObject *ref;

    switch (rule->getRuleType()) {

    case NATRule::SNAT:
    case NATRule::Masq:
    {
/*
 * one user reported that this code was not working right in his
 * configuration. He had to "external" interfaces on the firewall,
 * with overlapping subnets. DMZ subner had longer netmask overlapping
 * with external subnet, although default gateway was pointing to the
 * external interface. He needed to translate packets going from
 * inside NOT to DMz to the ip address that belonged to external
 * interface (there were few addresses on that interface, too).  This
 * code used to choose interface based on the address in TSrc, but
 * since his subnets overlapped, code picked DMZ interface. Choosing
 * interface (or rather address) with longer netmask did not help
 * since it was DMZ that had longer netmask in his configuration.
 *
 * I can't drop "-o" though. the following rule needs it:
 *
 * OSrc           ODst      TSrc          TDst
 * internal_net   any       fw            original
 *
 *  assuming fw has two external interfaces.
 * 
 * Here we should end up with two SNAT rules that translate the same
 * source address into two different addresses, depending on what
 * interface packet is exiting the firewall through.
 *
 * I had to make algorithm more complex because of this. See below.
 */

        RuleElementTSrc *tsrcrel=rule->getTSrc();
        ref=tsrcrel->front();
        a=Address::cast(compiler->getCachedObj(ref->getStr("ref")));
        if(a==NULL)
            throw FWException(_("Broken TSrc in ")+rule->getLabel());

    }
    break;

    case NATRule::DNAT:
    case NATRule::Redirect:
/*
 * fixed bug #576598 (user reported a problem with configuration where
 * DNAT translation was provided using IP addresses on DMZ (interface
 * eth3, routable IP), but connection was actually coming from
 * outside.  Compiler used to add "-i eth3" to the generated code
 * which broke his DNAT rule because packets were actually entering
 * firewall through different interface.
 *
 * our code determined that his ODst address belongs to DMZ and
 * therefore picked eth3 (interface on DMZ) for "-i" option. The rule
 * did not get split onto two (one for outside interface and another
 * for DMZ, since both were marked as external) because he had
 * specific address in ODst, not firewall object. It seems the only way
 * to solve this problem is to avoid "-i interface_name" all together.
 *
 * Unfortunately this creates another problem. If outside interface is
 * dynamic, the rule must have "-i ..." option since there is no other
 * way to make matching specific to connections coming from outside
 * (since address is unspecified).
 */
    {
        RuleElementODst *odstrel=rule->getODst();
        ref=odstrel->front();
        a=Address::cast(compiler->getCachedObj(ref->getStr("ref")));
        if(a==NULL)
            throw FWException(_("Broken ODst in ")+rule->getLabel());

        if (Interface::isA(a) && Interface::cast(a)->isDyn()) 
        {
/* this is interface and it is dynamic, we must associate this rule with
 * interface since destination address is unknown
 */
        }
        else
            return true;
    }
    break;

    default: 
	return true;
    }

/*
 * First, we look for an interface which is a parent of tsrc
 *
 */
    list<FWObject*> l2=compiler->fw->getByType(Interface::TYPENAME);
    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
        Interface *iface=Interface::cast(*i);
        assert(iface);
/* I believe this case works if interface is dynamic (dynamic
 * interface won't get split onto child objects and will stay in
 * ODst/TSrc, so it will be found here when we compare its ID with IDs
 * of all interfaces
 */
        if (iface->getId()==a->getId() || a->isChildOf(iface))
        {
            rule->setInterfaceId( iface->getId() );
            return true;
        }
    }

/* 
 * ODst/TSrc is not an interface or an address of one of the
 * interfaces; we have guess now 
 *
 *
 * Things are more complex than I thought before : interfaces could be
 * connected to overlapping subnets ... I need to scan interfaces,
 * find all that sit on a subnet to which adderss 'a' belongs and then
 * pick the one with the longest netmask (using Netmask::getLength() )
 */
    std::list<subnetInfo> sl;

    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	Interface *iface=Interface::cast(*i);
	assert(iface);

        if ( ! iface->isDyn()) 
        {
            list<FWObject*> l3=iface->getByType(IPv4::TYPENAME);
            for (list<FWObject*>::iterator i=l3.begin(); i!=l3.end(); ++i) 
            {
                IPv4 *ipv4=IPv4::cast(*i);
                assert(ipv4);

                IPNetwork n( ipv4->getAddress() , ipv4->getNetmask() );
                if ( n.belongs( a->getAddress() ) ) {  // && iface->isExt() ) {
                    sl.push_back( subnetInfo(iface,ipv4,ipv4->getNetmask().getLength()) );
                }
            }
        }
    }
    if (!sl.empty())
    {
        subnetInfo  sni;
        for (list<subnetInfo>::iterator i=sl.begin(); i!=sl.end(); i++)
            if (i->nmlength > sni.nmlength) { sni= (*i); }

        rule->setInterfaceId( sni.iface->getId() );
    }
    return true;
}


void NATCompiler_ipt::compile()
{
    FWOptions* options=fw->getOptionsObject();
    bool manage_virtual_addr=options->getBool("manage_virtual_addr");


    cout << _(" Compiling NAT rules for ") << fw->getName() << " ..." << endl << flush;

    try {

	Compiler::compile();

        add( new NATCompiler::Begin(                "Begin processing"                   ) );
        add( new printTotalNumberOfRules(                                                ) );

        add( new emptyGroupsInOSrc(          "check for empty groups in OSRC"         ) );
        add( new emptyGroupsInODst(          "check for empty groups in ODST"         ) );
        add( new emptyGroupsInOSrv(          "check for empty groups in OSRV"         ) );

        add( new emptyGroupsInTSrc(          "check for empty groups in TSRC"         ) );
        add( new emptyGroupsInTDst(          "check for empty groups in TDST"         ) );
        add( new emptyGroupsInTSrv(          "check for empty groups in TSRV"         ) );


	add( new ExpandGroups(                      "Expand groups"                      ) );
        add( new eliminateDuplicatesInOSRC(  "eliminate duplicates in OSRC"              ) );
        add( new eliminateDuplicatesInODST(  "eliminate duplicates in ODST"              ) );
        add( new eliminateDuplicatesInOSRV(  "eliminate duplicates in OSRV"              ) );

        add( new classifyNATRule(                   "decide on the rule type"            ) );
        add( new VerifyRules(                       "verify rules"                       ) );
//        add( new fillTranslatedSrv(                 "fill translated srv if it is any"   ) );
//	if ( manage_virtual_addr )
//            add( new addVirtualAddress(             "add virtual addresses"              ) );

        add( new portTranslationRules(              "port translation rules"             ) );
        add( new specialCaseWithRedirect(           "check for special case with redirecting port translation rules" ) );

        add( new doOSrcNegation(                    "process negation in OSrc"           ) );
        add( new doODstNegation(                    "process negation in ODst"           ) );
        add( new doOSrvNegation(                    "process negation in OSrv"           ) );

        add( new splitIfOSrcMatchesFw(              "split rule if OSrc matches FW"      ) );
        add( new splitNONATRule(                    "process NAT rules that request no translation"                ) );
        add( new decideOnChain(                     "decide on chain"                    ) );
        add( new decideOnTarget(                    "decide on target"                   ) );
        add( new ReplaceFirewallObjectsODst(        "replace firewall in ODst"           ) );
	add( new ReplaceFirewallObjectsTSrc(        "replace firewall in TSrc"           ) );
        add( new splitOnDynamicInterfaceInODst(     "split rule if ODst is dynamic interface" ) );
        add( new splitOnDynamicInterfaceInTSrc(     "split rule if TSrc is dynamic interface" ) );

        add( new ExpandMultipleAddresses(           "expand multiple addresses"          ) );
        add( new specialCaseWithDynInterface(       "check for a special cases with dynamic interface" ) );
        add( new addressRanges(                     "expand address ranges"              ) );

        add( new splitServices(                     "split on services"                  ) );
        add( new VerifyRules2(                      "check correctness of TSrv"          ) );
        add( new separatePortRanges(                "separate port ranges"               ) );
        add( new separateSourcePorts(               "separate objects with src"          ) );
        add( new separateSourceAndDestinationPorts( "separate objects with both src and dest ports" ) );
        add( new prepareForMultiport(               "prepare for multiport"              ) );
        add( new ConvertToAtomicForAddresses(       "convert to atomic rules"            ) );

	if ( manage_virtual_addr )
            add( new addVirtualAddress(             "add virtual addresses"              ) );

        add( new AssignInterface(                   "assign rules to interfaces"         ) );
        add( new dynamicInterfaceInODst(            "split if dynamic interface in ODst" ) );
	add( new dynamicInterfaceInTSrc(            "split if dynamic interface in TSrc" ) );
        add( new PrintRule(                         "generate iptables code"             ) );
        add( new simplePrintProgress(                                                    ) );

        runRuleProcessors();


    } catch (FWException &ex) {
	error(ex.toString());
	exit(1);
    }
}


void NATCompiler_ipt::epilog()
{
    try {

	output << "#\n#\n";

        output << endl;
        output << endl;

	cout << _(" NAT rules compiled successfully \n") << flush;

    } catch (FWException ex) {
	error(ex.toString());
	exit(1);
    }
}
