/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler_ipt.cc,v 1.71 2002/12/13 08:05:09 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "PolicyCompiler_ipt.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/CustomService.hh"
#include "fwbuilder/Policy.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/IPv4.hh"
#include "fwbuilder/Firewall.hh"

#include <stack>

#include <iostream>
#if __GNUC__ > 3 || \
    (__GNUC__ == 3 && (__GNUC_MINOR__ > 2 || (__GNUC_MINOR__ == 2 ) ) )
#  include <streambuf>
#else
#  include <streambuf.h>
#endif
#include <iomanip>
#include <fstream>
#include <sstream>

#include <assert.h>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

static int chain_no=0;

string PolicyCompiler_ipt::myPlatformName() { return "iptables"; }

string PolicyCompiler_ipt::getNewTmpChainName(PolicyRule *rule)
{
    std::ostringstream str;

    str << "ptmp" << setw(3) << setfill('0') << chain_no;
    chain_no++;

    return str.str();
}

string PolicyCompiler_ipt::getNewChainName(PolicyRule *rule,Interface *rule_iface)
{
    std::ostringstream str;

    if (rule_iface)  str << rule_iface->getName() << "_";

    switch (rule->getDirection()) {
    case PolicyRule::Inbound:   str << "In_";  break;
    case PolicyRule::Outbound:  str << "Out_"; break;
    }
    int pos=rule->getPosition();
    if (pos>=0)
        str << "RULE_" << pos;
    else // special case: position == -1
        str << "RULE_000";

    string suffix=rule->getStr("subrule_suffix");
    if (!suffix.empty()) str << "_" << suffix;

    chain_no++;

    return str.str();
}

string PolicyCompiler_ipt::getActionOnReject(PolicyRule *rule)
{
    FWOptions  *ruleopt =rule->getOptionsObject();
    return ruleopt->getStr("action_on_reject");
}

bool   PolicyCompiler_ipt::isActionOnRejectTCPRST(PolicyRule *rule)
{
    string s=getActionOnReject(rule);
    return ( ! s.empty() && s.find("TCP ")!=string::npos );
}

/*
 * resets rule option "action_on_reject" so it won't be TCP RST
 * Algorithm:
 *
 *    if global option "action_on_reject" is not empty
 *      if global option is TCP RST
 *        set rule option value to "none"
 *      else
 *        copy value from global option to rule option
 *    else
 *      set rule option value to "none"
 *
 * 
 */
void PolicyCompiler_ipt::resetActionOnReject(PolicyRule *rule)
{
    FWOptions  *ruleopt =rule->getOptionsObject();            
    string go=getCachedFwOpt()->getStr("action_on_reject");

    if (!go.empty())
    {
        if ( go.find("TCP ")!=string::npos )
        {
            ruleopt->setStr("action_on_reject","NOP"); // hack.
        } else
        {
            ruleopt->setStr("action_on_reject",go);
        }
    } else
        ruleopt->setStr("action_on_reject","none"); // hack.
}

int PolicyCompiler_ipt::prolog()
{
    if (fw->getStr("platform")!="iptables") 
	abort(_("Unsupported platform ") + fw->getStr("platform") );

    int n= PolicyCompiler::prolog();

    Service     *anytcp, *anyudp, *anyicmp, *anyip;
    TCPService  *tcpsyn;

    anytcp=Service::cast(dbcopy->create(TCPService::TYPENAME,false) );
    anytcp->setId(ANY_TCP_OBJ_ID);
    anytcp->setName("AnyTCP");
    dbcopy->add(anytcp);
    cacheObj(anytcp); // to keep cache consistent

    tcpsyn=TCPService::cast(dbcopy->create(TCPService::TYPENAME,false) );
    tcpsyn->setId(TCP_SYN_OBJ_ID);
    tcpsyn->setName("tcpSYN");
    tcpsyn->setTCPFlag(TCPService::SYN,true);
    tcpsyn->setAllTCPFlagMasks();
    dbcopy->add(tcpsyn);
    cacheObj(tcpsyn); // to keep cache consistent

    anyudp=Service::cast(dbcopy->create(UDPService::TYPENAME,false) );
    anyudp->setId(ANY_UDP_OBJ_ID);
    anyudp->setName("AnyUDP");
    dbcopy->add(anyudp);
    cacheObj(anyudp); // to keep cache consistent

    anyicmp=Service::cast(dbcopy->create(ICMPService::TYPENAME,false) );
    anyicmp->setId(ANY_ICMP_OBJ_ID);
    anyicmp->setName("AnyICMP");
    dbcopy->add(anyicmp);
    cacheObj(anyicmp); // to keep cache consistent

    anyip=Service::cast(dbcopy->create(IPService::TYPENAME,false) );
    anyip->setId(ANY_IP_OBJ_ID);
    anyip->setName("AnyIP");
    dbcopy->add(anyip);
    cacheObj(anyip); // to keep cache consistent

    return n;
}


bool PolicyCompiler_ipt::Logging1::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    if ( compiler->getCachedFwOpt()->getBool("log_all") )
        rule->setLogging(true);

    tmp_queue.push_back(rule);

    return true;
}


/**
 * splits rule if logging is required and either src or dst is
 * not any
 */
bool PolicyCompiler_ipt::Logging2::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule *rule=getNext(); if (rule==NULL) return false;
    Interface *rule_iface = compiler->getCachedFwInterface(rule->getInterfaceId());

    RuleElementSrc *nsrc;
    RuleElementDst *ndst;
    RuleElementSrv *nsrv;

    if (rule->getLogging()) {
	string new_chain=PolicyCompiler_ipt::getNewChainName(rule,rule_iface);

	PolicyRule *r;
        FWOptions  *ruleopt;

/*
 * add copy of original rule, but turn off logging and set target
 * chain to new_chain.
 */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        ruleopt =r->getOptionsObject();
	r->setStr("ipt_target",new_chain);
	r->setLogging(false);
        ruleopt->setInt("limit_value",-1);
	tmp_queue.push_back(r);

/*
 * need to add two rules with the same rule_label, direction=both, no
 * interface, no src, no srv and no dst. One of these new rules should
 * have target = LOG and another should inherit action and therefore
 * target from original rule. Both new rules go into chain new_chain.
 * In both rules turn off stateful inspection.
 *
 * keep interface information in the first one to be able to process
 * %I in log prefix
 *
 */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        ruleopt =r->getOptionsObject();
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("ipt_target","LOG");
	r->setDirection( PolicyRule::Both );
	r->setLogging(false);
        ruleopt->setBool("stateless",true);
 	tmp_queue.push_back(r);

	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        ruleopt =r->getOptionsObject();
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
/*
 * special case: need to preserve information about service protocol in case
 * action_on_reject is TCP RST
 */
        nsrv=r->getSrv();
        Service  *srv= nsrv->getFirst(true);
        if (TCPService::isA(srv))
        {
            nsrv->clearChildren();
            nsrv->addRef(compiler->getCachedObj(ANY_TCP_OBJ_ID));
        }
        else
        {
            nsrv->clearChildren();
            nsrv->setAnyElement();
        }

	r->setStr("ipt_chain",new_chain);
	r->setInterfaceId("nil");
	r->setDirection( PolicyRule::Both );
	r->setLogging(false);
        ruleopt->setBool("stateless",true);
        r->setBool("final",true);

	tmp_queue.push_back(r);
    } else
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::singleSrcNegation::processNext()
{
    PolicyRule         *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  

/*   ! A  B  C  ACTION  */
    if (srcrel->getNeg() && srcrel->size()==1 && src!=NULL &&
        !compiler->complexMatch(src,compiler->fw)) 
    {
        srcrel->setNeg(false);
        srcrel->setBool("single_object_negation",true);
    }

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::singleDstNegation::processNext()
{
    PolicyRule         *rule=getNext(); if (rule==NULL) return false;

    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  

/*   A  ! B  C  ACTION  */
    if (dstrel->getNeg() && dstrel->size()==1 && dst!=NULL &&
        !compiler->complexMatch(dst,compiler->fw)) 
    {
        dstrel->setNeg(false);
        dstrel->setBool("single_object_negation",true);
    }

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::singleSrvNegation::processNext()
{
    PolicyRule         *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srvrel=rule->getSrv();
    Service        *srv=srvrel->getFirst(true); // need to make sure it is not a group

/*   A  B  ! C  ACTION  */
    if (srvrel->getNeg() && srvrel->size()==1 && srv!=NULL ) 
    {
        srvrel->setNeg(false);
        srvrel->setBool("single_object_negation",true);
    }

    tmp_queue.push_back(rule);
    return true;
}


bool PolicyCompiler_ipt::SrcNegation::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule         *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();

/*   ! A  B  C  ACTION  */

    if (srcrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;
        FWOptions      *ruleopt;

	string new_chain=PolicyCompiler_ipt::getNewTmpChainName(rule);
	srcrel->setNeg(false);


/*     any  B C  TMP_CHAIN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","1");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);

	tmp_queue.push_back(r);

/* TMP_CHAIN   A  any any  RETURN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","2");
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);
//	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","3");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
/*
 * special case: need to preserve information about service protocol in case
 * action_on_reject is TCP RST
 */
        nsrv=r->getSrv();
        Service  *srv= nsrv->getFirst(true);
        if (TCPService::isA(srv))
        {
            nsrv->clearChildren();
            nsrv->addRef(compiler->getCachedObj(ANY_TCP_OBJ_ID));
        }
        else
        {
            nsrv->clearChildren();
            nsrv->setAnyElement();
        }
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setInterfaceId("nil");
        r->setBool("final",true);
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::DstNegation::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementDst *dstrel=rule->getDst();

/*   A  ! B  C  ACTION  */

    if (dstrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;
        FWOptions      *ruleopt;

	string new_chain=PolicyCompiler_ipt::getNewTmpChainName(rule);
	dstrel->setNeg(false);


/*     A any C  TMP_CHAIN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","1");
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);
	tmp_queue.push_back(r);

/* TMP_CHAIN   any B any  RETURN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","2");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);
//	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","3");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
/*
 * special case: need to preserve information about service protocol in case
 * action_on_reject is TCP RST
 */
        nsrv=r->getSrv();
        Service  *srv= nsrv->getFirst(true);
        if (TCPService::isA(srv))
        {
            nsrv->clearChildren();
            nsrv->addRef(compiler->getCachedObj(ANY_TCP_OBJ_ID));
        }
        else
        {
            nsrv->clearChildren();
            nsrv->setAnyElement();
        }
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setInterfaceId("nil");
        r->setBool("final",true);
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::SrvNegation::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srvrel=rule->getSrv();

/*   A  B  !C  ACTION  */

    if (srvrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;
        FWOptions      *ruleopt;

	string new_chain=PolicyCompiler_ipt::getNewTmpChainName(rule);
	srvrel->setNeg(false);


/*     A B any  TMP_CHAIN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","1");
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  any  C  RETURN  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","2");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
        ruleopt =r->getOptionsObject();
        ruleopt->setInt("limit_value",-1);
        ruleopt->setInt("limit_value",-1);
//	r->setInterfaceId("nil");
	tmp_queue.push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
        r->setStr("subrule_suffix","3");
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
        nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setInterfaceId("nil");
        r->setBool("final",true);
	tmp_queue.push_back(r);

    } else
	tmp_queue.push_back(rule);

    return true;
}









bool PolicyCompiler_ipt::InterfaceAndDirection::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    Interface *rule_iface = compiler->getCachedFwInterface(rule->getInterfaceId());
    
    if (rule->getDirection()==PolicyRule::Undefined) 
	rule->setDirection( PolicyRule::Both );

    if (rule_iface==NULL && rule->getDirection()==PolicyRule::Both) {
	rule->setInterfaceId("nil");
	return true;
    }

    if (rule_iface==NULL && (
	    rule->getDirection()==PolicyRule::Inbound ||
	    rule->getDirection()==PolicyRule::Outbound )
    ) throw FWException(_("Direction set without interface in rule ")+rule->getLabel());
}

bool PolicyCompiler_ipt::splitIfIfaceAndDirectionBoth::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    Interface *rule_iface = compiler->getCachedFwInterface(rule->getInterfaceId());

    if (rule_iface!=NULL && rule->getDirection()==PolicyRule::Both) {
	PolicyRule *r;

	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Inbound );
	tmp_queue.push_back(r);

	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Outbound );
	tmp_queue.push_back(r);
    } else
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::bridgingFw::checkForMatchingBroadcastAndMulticast(Address *addr)
{

    IPAddress obj1_addr=addr->getAddress();
    if (obj1_addr!=IPAddress("0.0.0.0") && 
        (obj1_addr.isBroadcast() || obj1_addr.isMulticast())
    ) return true;

    FWObjectTypedChildIterator j=compiler->fw->findByType(Interface::TYPENAME);
    for ( ; j!=j.end(); ++j ) {
        Interface *iface=Interface::cast(*j);
        if ( ! iface->isDyn() ) {

            FWObjectTypedChildIterator k=iface->findByType(IPv4::TYPENAME);
            for ( ; k!=k.end(); ++k ) {
                IPv4 *ipv4=IPv4::cast(*k);

/*
 * commented out to fix bug #637694 - "bridge enbaled / management"
 * Rule where firewall was in destination, and bridging option was on,
 * yielded code in FORWARD chain when this line was uncommented.  

   if ( ipv4->getAddress()==obj1_addr ) return true;

 */
                IPNetwork n( ipv4->getAddress() , ipv4->getNetmask() );
                if (n.getBroadcastAddress()==obj1_addr)  return true;
            }
        }
    }
    return false;
}

/*
 * call this after splitIfSrcMatchesFw and splitIfDstMatchesFw so that
 * we can count on firewall or broadcast/multicast being a single
 * object in src and dst.
 */
bool PolicyCompiler_ipt::bridgingFw::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    Address        *src=rule->getSrc()->getFirst(true);
    Address        *dst=rule->getDst()->getFirst(true);


    if ( rule->getStr("ipt_chain")=="OUTPUT" || rule->getStr("ipt_chain")=="INPUT" )
    {

        if (checkForMatchingBroadcastAndMulticast(src) ||
            checkForMatchingBroadcastAndMulticast(dst) )
            rule->setStr("ipt_chain","FORWARD");
    }


    tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::splitIfSrcAny::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    if ( ! rule->getStr("ipt_chain").empty() ) {
	tmp_queue.push_back(rule);
	return true;
    }

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src=srcrel->getFirst(true);

    if ( rule->getDirection()!=PolicyRule::Inbound &&
         ( 
             srcrel->isAny()  ||

             ( srcrel->size()==1 && src!=NULL &&
               !compiler->complexMatch(src,compiler->fw) &&
               srcrel->getBool("single_object_negation")) 
         ) 
    )
    {
         
        PolicyRule *r= PolicyRule::cast(
            compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
        compiler->temp_ruleset->add(r);
        r->duplicate(rule);
        r->setStr("ipt_chain","OUTPUT");
        r->setDirection( PolicyRule::Outbound );
        tmp_queue.push_back(r);
    }
    tmp_queue.push_back(rule);  // add old rule anyway

    return true;
}

bool PolicyCompiler_ipt::splitIfDstAny::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    if ( ! rule->getStr("ipt_chain").empty() ) {
	tmp_queue.push_back(rule);
	return true;
    }

    RuleElementDst *dstrel=rule->getDst();
    Address        *dst=dstrel->getFirst(true);

    if ( rule->getDirection()!=PolicyRule::Outbound &&
         ( 
             dstrel->isAny()  ||

             ( dstrel->size()==1 && dst!=NULL &&
               !compiler->complexMatch(dst,compiler->fw) &&
               dstrel->getBool("single_object_negation")) 
         ) 
    )
    {
	PolicyRule *r= PolicyRule::cast(
	    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setStr("ipt_chain","INPUT");
	r->setDirection( PolicyRule::Inbound );
	tmp_queue.push_back(r);
    }
    tmp_queue.push_back(rule); // add old rule in any case

    return true;
}

/*
 *  predicates that run before guarantee that when we call this one,
 *  firewall object, if it is in src or dst, is single object there
 */
bool PolicyCompiler_ipt::checkSrcAndDst1::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if (src->getId()!=compiler->getFwId() &&
	dst->getId()==compiler->getFwId() &&
	rule->getDirection()==PolicyRule::Outbound )
	throw FWException(_("direction can not be outbound when destination is firewall, in rule ")+rule->getLabel());

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::checkSrcAndDst2::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if (src->getId()==compiler->getFwId() &&
	dst->getId()!=compiler->getFwId() &&
	rule->getDirection()==PolicyRule::Inbound )
	throw FWException(_("direction can not be inbound when source is firewall, in rule ")+rule->getLabel());

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::specialCaseWithFW1::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  
    if(src==NULL) throw(_("Broken SRC in rule ")+rule->getLabel());
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);
    if(dst==NULL) throw(_("Broken DST in rule ")+rule->getLabel());

    if (!src->isAny() && !dst->isAny() &&
        compiler->complexMatch(src,compiler->fw) &&
        compiler->complexMatch(dst,compiler->fw) &&
	rule->getDirection()== PolicyRule::Both ) {

	PolicyRule *r;

	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Inbound );
	tmp_queue.push_back(r);

	r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Outbound );
	tmp_queue.push_back(r);
    } else
	tmp_queue.push_back(rule);

    return true;
}

void PolicyCompiler_ipt::specialCaseWithDynInterface::dropDynamicInterface(RuleElement *re)
{
    list<FWObject*> cl;
    for (list<FWObject*>::iterator i1=re->begin(); i1!=re->end(); ++i1) 
    {
        FWObject *o   = *i1;
        FWObject *obj = o;
        if (FWReference::cast(o)!=NULL) obj=compiler->getCachedObj(o->getStr("ref"));
        Interface  *ifs   =Interface::cast( obj );

        if (ifs!=NULL && ifs->isDyn()) continue;
        cl.push_back(obj);
    }
    if (!cl.empty()) 
    {
        re->clearChildren();
        for (list<FWObject*>::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  
            re->addRef( (*i1) );
    }
}

/**
 * checks for the following situations: 
 *
 *  1. dynamic interface is in source and direction is inbound
 *  (drop interface from src since source address is
 *  undertermined)
 *
 *  2. dynamic interface is in source, direction is outbound
 *  and chain is  temporary (drop interface from the list, this
 *  rule has been created while processing negation. TODO: this
 *  is kludge, need to create separate temporary chain while
 *  doing negation in src if one of the objects is firewall)
 *
 *  3. dynamic interface is in destination and chain is "OUTPUT"
 *  (drop interface since dest. address is undefined)
 * 
 */
bool PolicyCompiler_ipt::specialCaseWithDynInterface::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    switch ( rule->getDirection() ) 
    {
    case PolicyRule::Inbound:
        dropDynamicInterface( rule->getSrc() );
        break;
    case PolicyRule::Outbound:
        if ( rule->getStr("ipt_chain")=="OUTPUT" ) 
            dropDynamicInterface( rule->getDst() );
        else
            dropDynamicInterface( rule->getSrc() );
        break;
    }

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::expandMultipleAddressesIfNotFW::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if (Firewall::cast(src)==NULL) 
        compiler->_expandAddr(rule,srcrel);

    if (Firewall::cast(dst)==NULL) 
        compiler->_expandAddr(rule,dstrel);


    tmp_queue.push_back(rule);

    return true;
}


bool PolicyCompiler_ipt::decideOnChain::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);

    tmp_queue.push_back(rule);

    if ( ! rule->getStr("ipt_chain").empty() ) return true;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    rule->setStr("ipt_chain","FORWARD");

//    if ( compiler->getCachedFwOpt()->getBool("bridging_fw") ) return true;
    bool b,m;
/* 
 * do not check for broadcasts and multicasts in bridging firewall because
 * they should go to FORWARD chain 
 */
    b=m= !( compiler->getCachedFwOpt()->getBool("bridging_fw") );

    switch ( rule->getDirection() ) {

    case PolicyRule::Inbound:
/* if direction is "Inbound", chain can never be OUTPUT, but could be FORWARD */
        if (!dst->isAny() && ipt_comp->complexMatch(dst,ipt_comp->fw,b,m))
            rule->setStr("ipt_chain","INPUT");
        return true;


    case PolicyRule::Outbound:
/* if direction is "Outbound", chain can never be INPUT, but could be FORWARD */
        if (!src->isAny() && ipt_comp->complexMatch(src,ipt_comp->fw,b,m))
            rule->setStr("ipt_chain","OUTPUT");
	return true;

    default:
/* direction == Both
 *
 * check if src or dst is Interface with no IP address but with MAC address.
 * TODO: define object MACAddress
 */
        Interface *iface;

        bool src_mac_only=false;
        if ( (iface=Interface::cast(src))!=NULL ) {
            src_mac_only=( iface->getIPAddress().to32BitInt()==0  && iface->getPhysicalAddress()!="" );
        }

        bool dst_mac_only=false;
        if ( (iface=Interface::cast(dst))!=NULL ) {
            dst_mac_only=( iface->getIPAddress().to32BitInt()==0  && iface->getPhysicalAddress()!="" );
        }


        if (!dst->isAny() && !dst_mac_only && ipt_comp->complexMatch(dst,ipt_comp->fw,b,m)) {
            rule->setStr("ipt_chain","INPUT");
            return true;
        }
        if (!src->isAny() && !src_mac_only && ipt_comp->complexMatch(src,ipt_comp->fw,b,m)) {
            rule->setStr("ipt_chain","OUTPUT");
            return true;
        }
    }
}

bool PolicyCompiler_ipt::decideOnTarget::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    if ( ! rule->getStr("ipt_target").empty() ) return true; // already defined

    switch (rule->getAction()) {
    case PolicyRule::Accept:  rule->setStr("ipt_target","ACCEPT"); break;
    case PolicyRule::Deny:    rule->setStr("ipt_target","DROP");   break;
    case PolicyRule::Reject:  rule->setStr("ipt_target","REJECT"); break;
    case PolicyRule::Return:  rule->setStr("ipt_target","RETURN"); break;
    }
    return true;
}

/*
 *  TODO:  may be make this optional, with checkbox in Firewall tab
 */
bool PolicyCompiler_ipt::removeFW::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);


    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if ( rule->getStr("ipt_chain")=="INPUT" && 
	 dst->getId()==compiler->getFwId() ) {
	dstrel->clearChildren();
	dstrel->setAnyElement();
    }

    if ( rule->getStr("ipt_chain")=="OUTPUT" && 
	 src->getId()==compiler->getFwId() ) {
	srcrel->clearChildren();
	srcrel->setAnyElement();
    }
    return true;
}

bool PolicyCompiler_ipt::separatePortRanges::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *rel= rule->getSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( TCPService::isA(s) || UDPService::isA(s) ) {
            int srs=s->getInt("src_range_start");
            int sre=s->getInt("src_range_end");
            int drs=s->getInt("dst_range_start");
            int dre=s->getInt("dst_range_end");

            compiler->normalizePortRange(srs,sre);
            compiler->normalizePortRange(drs,dre);

            if (srs!=sre || drs!=dre) {
                PolicyRule *r= PolicyRule::cast(
                    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
                compiler->temp_ruleset->add(r);
                r->duplicate(rule);
                RuleElementSrv *nsrv=r->getSrv();
                nsrv->clearChildren();
                nsrv->addRef( s );
                tmp_queue.push_back(r);
                services.push_back(s);
            } 
        }
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::separateSrcPort::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *rel= rule->getSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( TCPService::isA(s) || UDPService::isA(s) ) {
            int srs=s->getInt("src_range_start");
            int sre=s->getInt("src_range_end");

            compiler->normalizePortRange(srs,sre);

            if (srs!=0 || sre!=0) {
                PolicyRule *r= PolicyRule::cast(
                    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
                compiler->temp_ruleset->add(r);
                r->duplicate(rule);
                RuleElementSrv *nsrv=r->getSrv();
                nsrv->clearChildren();
                nsrv->addRef( s );
                tmp_queue.push_back(r);
                services.push_back(s);
            } 
        }
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::fillActionOnReject::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    FWOptions  *ruleopt =rule->getOptionsObject();
    string s=ruleopt->getStr("action_on_reject");
    if (s.empty()) ruleopt->setStr("action_on_reject",
                     ipt_comp->getCachedFwOpt()->getStr("action_on_reject"));

    tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::expandGroupsInSrv::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srv= rule->getSrv();

    compiler->expandGroupsInRuleElement(srv);

    tmp_queue.push_back(rule);
    return true;
}

/*
 * I am adding subrule suffix here, which I then use to generate
 * unique new chain name for this rule. The idea is to generate
 * meaningful chain name, which is associated with rule number (like
 * RULE_5), however since this processor runs very early and may split
 * the rule, subsequent processors that create new chains end up
 * creating chains with the same names. Need this suffix to create
 * different, yet meaningful chain names.
 *
 * TODO: add methods addRuleSuffix and getRuleSuffix to class Rule.
 * Define suffix automatically in a tree-like manner, so that when
 * original rule is split, its parts will get suffixes ".1" and
 * ".2". When these parts are split again, suffixes should become
 * ".1.1" and ".1.2" and so on.
 */
bool PolicyCompiler_ipt::splitServicesIfRejectWithTCPReset::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srv= rule->getSrv();

    if ( rule->getAction()==PolicyRule::Reject && ipt_comp->isActionOnRejectTCPRST(rule))
    {
        list<Service*> tcp;
        list<Service*> other;

        for (FWObject::iterator i=srv->begin(); i!=srv->end(); ++i) 
        {
            FWObject *o= *i;
            if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	    
            Service *s1=Service::cast( o );
            assert(s1);

            if ( TCPService::isA(s1) ) tcp.push_back(s1);
            else                       other.push_back(s1);
        }

        if ( !other.empty() && tcp.empty() )
        {
            if (seen_rules[rule->getPosition()]==false) 
                compiler->warning(_("Rule action 'Reject' with TCP RST can be used only with TCP services. Rule ")+rule->getLabel());
            ipt_comp->resetActionOnReject(rule);
            tmp_queue.push_back(rule);
            seen_rules[rule->getPosition()]=true;
            return true;
        }

        if ( other.empty() && !tcp.empty() )
        {
            tmp_queue.push_back(rule);
            return true;
        }
/* if both are not empty */

        PolicyRule *r;
        RuleElementSrv *nsrv;

        r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
        compiler->temp_ruleset->add(r);
        r->duplicate(rule);
        nsrv=r->getSrv();
        nsrv->clearChildren();

        for (list<Service*>::iterator j=other.begin(); j!=other.end(); j++)
            nsrv->addRef( (*j) );

        r->getOptionsObject()->setStr("action_on_reject","");
        r->setStr("subrule_suffix","1");
        tmp_queue.push_back(r);

        r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
        compiler->temp_ruleset->add(r);
        r->duplicate(rule);
        nsrv=r->getSrv();
        nsrv->clearChildren();

        for (list<Service*>::iterator j=tcp.begin(); j!=tcp.end(); j++)
            nsrv->addRef( (*j) );

        r->setStr("subrule_suffix","2");
        tmp_queue.push_back(r);
        return true;
    }
    tmp_queue.push_back(rule);

    return true;
}


/*
 * processor splitServices should have been called eariler, so now all
 * services in Srv are of the same type
 */
bool PolicyCompiler_ipt::prepareForMultiport::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *rel= rule->getSrv();
    Service        *srv= rel->getFirst(true);

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    if (IPService::isA(srv) || ICMPService::isA(srv) || CustomService::isA(srv)) {
/* multiport does  not support ip and icmp services, split the rule */
	for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	    
	    Service *s=Service::cast( o );
	    assert(s);

	    PolicyRule *r= PolicyRule::cast(
		compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	    compiler->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementSrv *nsrv=r->getSrv();
	    nsrv->clearChildren();
	    nsrv->addRef( s );

	    tmp_queue.push_back(r);
	}
	return true;
    }

    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	rule->setBool("ipt_multiport",true);

/* make sure we have no more than 15 ports */
	if (rel->size()>15) {

	    int n=0;
	    PolicyRule *r;
	    RuleElementSrv *nsrv;
	    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {

		FWObject *o= *i;
		if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	    
		Service *s=Service::cast( o );
		assert(s);

		if (n==0) {
		    r= PolicyRule::cast(
			compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
		    compiler->temp_ruleset->add(r);
		    r->duplicate(rule);
		    nsrv=r->getSrv();
		    nsrv->clearChildren();
		    tmp_queue.push_back(r);
		}
		nsrv->addRef( s );
		if (++n>=15) n=0;
	    }

	} else {
	    tmp_queue.push_back(rule);
	}
    }

    return true;
}

/*
 *  processor splitServices should have been called before, it makes sure
 *  all objects in Service are of the same type.
 *
 *  One special case is custom service "ESTABLISHED". This processor
 *  splits rule if it finds this service and turns off stateful
 *  inspection on the rule.
 *
 */
bool PolicyCompiler_ipt::specialCasesWithCustomServices::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srv= rule->getSrv();

    if (srv->isAny()) {
        tmp_queue.push_back(rule);
        return true;
    }

    stack<FWObject*> cl;
    for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) 
    {
        FWObject *o= *i;
        if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
        assert(o!=NULL);
        if (CustomService::isA(o)) 
        {
            string code=CustomService::cast(o)->getCodeForPlatform(compiler->myPlatformName());
            if (code.find("ESTABLISHED")!=string::npos ||
                code.find("RELATED")!=string::npos) 
            {
                PolicyRule *r= PolicyRule::cast(
                    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
                compiler->temp_ruleset->add(r);
                r->duplicate(rule);
                RuleElementSrv *nsrv=r->getSrv();
                nsrv->clearChildren();
                nsrv->addRef(o);
                r->getOptionsObject()->setBool("stateless",true);
                tmp_queue.push_back(r);

                cl.push(o);
            }
        }
    }
    while (!cl.empty()) {
        srv->removeRef( cl.top() );
        cl.pop();
    }
/*
 * if srv is 'any' at this point, then it had only single object at
 * the beginning and that object was CustomService which we've split
 * into a new rule. There is nothing left in the original srv so we
 * can simply drop the old rule.
 */
    if (!srv->isAny())
        tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::prepareForMACMatching::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

//    tmp_queue.push_back(rule);

    RuleElementSrc *rel= rule->getSrc();

    list<FWObject*> cl;
    for (FWObject::iterator i1=rel->begin(); i1!=rel->end(); ++i1) {
        FWObject *o= *i1;
        if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
        assert(o!=NULL);

	FWOptions *hopt;
	if (Host::isA(o) && (hopt=Host::cast(o)->getOptionsObject())!=NULL 
	    && hopt->getBool("use_mac_addr_filter") ) {
	    
	    cl.push_back(o);
	}
    }
    if ( ! cl.empty()) {
        for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  {
	    Host *h=Host::cast(*i1);
	    FWObjectTypedChildIterator j=h->findByType(Interface::TYPENAME);
	    if (j==j.end())
		throw FWException(_("MAC address match ordered for host object ")+
				  h->getName()+_(", but it has no interfaces. Rule ")+
				  rule->getLabel());
	    rel->removeRef(h);

	    bool got_at_least_one_mac=false;
	    for ( ; j!=j.end(); ++j ) {
		Interface *iface=Interface::cast(*j);
		if ( ! iface->getPhysicalAddress().empty() ) {
		    got_at_least_one_mac=true;

                    PolicyRule *r= PolicyRule::cast(
                        compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
                    compiler->temp_ruleset->add(r);
                    r->duplicate(rule);
		    r->setBool("use_mac_addr",true);
		    r->setStr("src_mac_addr" ,iface->getPhysicalAddress());

                    RuleElementSrc *nsrc= r->getSrc();
                    nsrc->clearChildren();
		    nsrc->addRef(iface);

                    tmp_queue.push_back(r);
		}
	    }
	    if ( ! got_at_least_one_mac )
		throw FWException(_("MAC address match ordered for host object ")+
				  h->getName()+_(", but none of its interfaces has MAC address defined. Rule ")+
				  rule->getLabel());
	}
        if (!rel->isAny()) tmp_queue.push_back(rule);
    } else
        tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::convertAnyToNotFW::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    RuleElementDst *dstrel=rule->getDst();

    if (srcrel->isAny())
    {
        srcrel->addRef(compiler->fw);
        srcrel->setNeg(true);
    }

    if (dstrel->isAny())
    {
        dstrel->addRef(compiler->fw);
        dstrel->setNeg(true);
    }

    tmp_queue.push_back(rule);
    return true;
}

bool PolicyCompiler_ipt::addPredefinedRules::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    if (add_once)
    {
        if ( ! compiler->getCachedFwOpt()->getBool("accept_new_tcp_with_no_syn") ) 
        {
            PolicyRule *r;
            RuleElementSrc *nsrc;
            RuleElementDst *ndst;
            RuleElementSrv *nsrv;

            r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
            compiler->temp_ruleset->add(r);
            r->setAction(PolicyRule::Deny);
            r->setLogging(true);
            r->setDirection(PolicyRule::Both);
            r->setPosition(-1);
            r->setHidden(true);
            r->setLabel("dropping TCP sessions opened prior firewall restart");

            nsrv=r->getSrv();    nsrv->clearChildren();  
            nsrv->addRef(compiler->getCachedObj(TCP_SYN_OBJ_ID));
            nsrv->setNeg(true);
            r->setBool("force_state_check",true);
            tmp_queue.push_back(r);

            if ( ! compiler->getCachedFwOpt()->getBool("firewall_is_part_of_any_and_networks") ) 
            {
                r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
                compiler->temp_ruleset->add(r);
                r->setAction(PolicyRule::Deny);
                r->setLogging(true);
                r->setDirection(PolicyRule::Both);
                r->setPosition(-1);
                r->setHidden(true);
                r->setLabel("dropping TCP sessions opened prior firewall restart");
                ndst=r->getDst();
                ndst->clearChildren();  
                ndst->addRef(compiler->fw );
                nsrv=r->getSrv();   nsrv->clearChildren();  
                nsrv->addRef(compiler->getCachedObj(TCP_SYN_OBJ_ID));
                nsrv->setNeg(true);
                r->setBool("force_state_check",true);
                tmp_queue.push_back(r);

                r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
                compiler->temp_ruleset->add(r);
                r->setAction(PolicyRule::Deny);
                r->setLogging(true);
                r->setDirection(PolicyRule::Both);
                r->setPosition(-1);
                r->setHidden(true);
                r->setLabel("dropping TCP sessions opened prior firewall restart");
                nsrc=r->getSrc();   nsrc->clearChildren();  nsrc->addRef(compiler->fw );
                nsrv=r->getSrv();   nsrv->clearChildren();  
                nsrv->addRef(compiler->getCachedObj(TCP_SYN_OBJ_ID));
                nsrv->setNeg(true);

                r->setBool("force_state_check",true);
                tmp_queue.push_back(r);
            }
        }
        add_once=false;
    } 

    tmp_queue.push_back(rule);

    return true;
}

/**
 *-----------------------------------------------------------------------
 */
void PolicyCompiler_ipt::compile()
{
    cout << _(" Compiling policy for ") << fw->getName() << " ..." <<  endl << flush;

    try {

	Compiler::compile();

        if ( fw->getOptionsObject()->getBool ("check_shading") ) 
        {
            add( new Begin                       (" Detecting rule shading"                ) );
            add( new printTotalNumberOfRules     (                                         ) );
            if ( ! fwopt->getBool("firewall_is_part_of_any_and_networks") )
                add( new convertAnyToNotFW       ("convert 'any' to '!fw'"                 ) );
            add( new ExpandGroups                ("expand groups"                          ) );
            add( new eliminateDuplicatesInSRC    ("eliminate duplicates in SRC"            ) );
            add( new eliminateDuplicatesInDST    ("eliminate duplicates in DST"            ) );
            add( new eliminateDuplicatesInSRV    ("eliminate duplicates in SRV"            ) );
            add( new ExpandMultipleAddressesInSRC("expand objects with multiple addresses in SRC" ) );
            add( new ExpandMultipleAddressesInDST("expand objects with multiple addresses in DST" ) );
            add( new ConvertToAtomic             ("convert to atomic rules"                ) );
            add( new DetectShading               ("Detect shading"                         ) );
            add( new simplePrintProgress         (                                         ) );

            runRuleProcessors();
            deleteRuleProcessors();
        }


        add( new PolicyCompiler::Begin(      "Begin processing"                      ) );
        add( new addPredefinedRules(         "Add some predefined rules"             ) );
        add( new printTotalNumberOfRules(                                            ) );
        add( new Logging1(                   "check global logging override option"  ) );
        add( new InterfaceAndDirection(      "fill in interface and direction"       ) );
        add( new splitIfIfaceAndDirectionBoth( "split interface rule with direction 'both'") );

        add( new emptyGroupsInSrc(           "check for empty groups in SRC"         ) );
        add( new emptyGroupsInDst(           "check for empty groups in DST"         ) );
        add( new emptyGroupsInSrv(           "check for empty groups in SRV"         ) );

        add( new singleSrvNegation(          "process negation in Srv if it holds single object"               ) );
        add( new SrvNegation(                "process negation in Srv"               ) );
        add( new expandGroupsInSrv(          "expand groups in Srv"                  ) );
        add( new fillActionOnReject(         "fill in action_on_reject"              ) );
        add( new splitServicesIfRejectWithTCPReset("check and split if action on reject is TCP reset"));
        add( new fillActionOnReject(         "fill in action_on_reject 2"            ) );
        add( new splitServicesIfRejectWithTCPReset("check and split if action on reject is TCP reset 2"));
        add( new singleSrcNegation(          "process negation in Src if it holds single object"               ) );
        add( new singleDstNegation(          "process negation in Dst if it holds single object"               ) );
        add( new SrcNegation(                "process negation in Src"               ) );
        add( new DstNegation(                "process negation in Dst"               ) );
        add( new Logging2(                   "process logging"                       ) );

	if ( ! fwopt->getBool("bridging_fw") &&
             fwopt->getBool("firewall_is_part_of_any_and_networks") ) {
            add( new splitIfSrcAny(          "split rule if src is any"              ) );
            add( new splitIfDstAny(          "split rule if dst is any"              ) );
        }


        add( new ExpandGroups(               "expand all groups"                     ) );
        add( new eliminateDuplicatesInSRC(   "eliminate duplicates in SRC"           ) );
        add( new eliminateDuplicatesInDST(   "eliminate duplicates in DST"           ) );
        add( new eliminateDuplicatesInSRV(   "eliminate duplicates in SRV"           ) );

        add( new splitIfSrcMatchesFw( "split rule if src matches FW") );
        add( new splitIfDstMatchesFw( "split rule if dst matches FW") );

/*  at this point in all rules where firewall is in either src or dst,
 *  firewall is single object in that rule element. Other rule
 *  elements may contain multiple objects yet
 */

        add( new specialCaseWithFW1(         "special case with firewall"            ) );
        add( new prepareForMACMatching(      "prepare for MAC filtering"             ) );

        add( new expandMultipleAddressesIfNotFW("expand multiple addresses if not FW") );
        add( new decideOnChain(              "decide on chain"                       ) );
        add( new decideOnTarget(             "decide on target"                      ) );
        add( new removeFW(                   "remove fw"                             ) );
        add( new ExpandMultipleAddresses(    "expand multiple addresses"             ) );

	if ( fwopt->getBool("bridging_fw") ) 
            add( new bridgingFw(                 "handle bridging firewall cases"        ) );

        add( new specialCaseWithDynInterface("check for a special cases with dynamic interface" ) );
        add( new addressRanges(              "process address ranges"                ) );
        add( new splitServices(              "split on services"                     ) );
//        add( new checkForTCPReset(           "check for action REJECT with TCP RST"  ) );
        add( new separateTCPWithFlags(       "split on TCP services with flags"      ) );
        add( new verifyCustomServices(       "verify custom services"                ) );
        add( new specialCasesWithCustomServices("scpecial cases with some custom services"      ) );
        add( new separatePortRanges(         "separate port ranges"                  ) );
        add( new separateSrcPort(            "split on TCP and UDP with source ports") );
        add( new optimize1(                  "optimization 1, pass 1"                ) );
        add( new optimize1(                  "optimization 1, pass 2"                ) );
        add( new optimize2(                  "optimization 2"                        ) );
        add( new prepareForMultiport(        "prepare for multiport"                 ) );
        add( new ConvertToAtomicForAddresses("convert to atomic rules by address elements") );
        add( new checkForZeroAddr(           "check for zero addresses"              ) );
        add( new ConvertToAtomicForIntervals("convert to atomic rules by interval element") );
        add( new PrintRule(                  "generate iptables code"                ) );
        add( new simplePrintProgress(                                                ) );

        runRuleProcessors();

    } catch (FWException &ex) {
	error(ex.toString());
	exit(1);
    }
}

string PolicyCompiler_ipt::debugPrintRule(Rule *r)
{
    PolicyRule *rule=PolicyRule::cast(r);
    FWOptions  *ruleopt =rule->getOptionsObject();
    string s= PolicyCompiler::debugPrintRule(rule)+
        " "+rule->getStr("ipt_chain")+" "+rule->getStr("ipt_target");
    if (rule->getAction()==PolicyRule::Reject)
        s=s+" "+ruleopt->getStr("action_on_reject");
    return s;
}

void PolicyCompiler_ipt::epilog()
{
    cout << _(" Policy compiled successfully \n") << flush;
}
